import json
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import cast

from pytyped.json.decoder import JsonDecoder
from pytyped.json.decoder import JsonNoneDecoder
from pytyped.json.encoder import AutoJsonEncoder
from pytyped.json.encoder import JsonBoxedEncoder
from pytyped.json.encoder import JsonEncoder
from pytyped.json.encoder import JsonEncoderException
from pytyped.json.encoder import JsonNoneEncoder
from pytyped.json.encoder import JsonTaggedEncoder
from tests.json import common
from tests.json.common import C1
from tests.json.common import G
from tests.json.common import G2

auto_json_encoder = AutoJsonEncoder()

a_encoder = auto_json_encoder.extract(common.A)
b_encoder = auto_json_encoder.extract(common.B)
boxed_a_encoder: JsonBoxedEncoder[common.A] = JsonBoxedEncoder("value", a_encoder)
c_flat_encoder = auto_json_encoder.extract(common.C)
c_nested_encoder = cast(
    JsonEncoder[common.C],
    JsonTaggedEncoder(
        branches={
            "C1": (common.C1, auto_json_encoder.extract(common.C1)),
            "C2": (common.C2, auto_json_encoder.extract(common.C2))
        },
        tag_field_name="tag",
        value_field_name="value"
    )
)
c_untagged_encoder = cast(JsonEncoder[common.C], auto_json_encoder.extract(cast(type, Union[common.C1, common.C2])))
str2int_dic_encoder = auto_json_encoder.extract(Dict[str, int])


test_cases: List[Tuple[JsonEncoder[Any], JsonDecoder[Any], List[str]]] = [
    (a_encoder, common.a_decoder, common.valid_a_jsons),
    (b_encoder, common.b_decoder, common.valid_b_jsons),
    (boxed_a_encoder, common.boxed_a_decoder, common.valid_a_jsons),
    (c_flat_encoder, common.c_flat_decoder, common.c_valid_flat_json_strs),
    (c_nested_encoder, common.c_nested_decoder, common.c_valid_nested_json_strs),
    (c_untagged_encoder, common.c_untagged_decoder, common.c_valid_untagged_json_strs),
    (str2int_dic_encoder, common.string_to_int_dic_json_decoder, common.valid_string_dictionary_jsons),
    (JsonNoneEncoder(), JsonNoneDecoder(), ['{}', 'null']),
    (auto_json_encoder.extract(common.Composite), common.composite_decoder, common.valid_composite_jsons)
]


def test_json_encoding() -> None:
    for (index, (encoder, decoder, json_strs)) in enumerate(test_cases):
        for json_str in json_strs:
            try:
                decoded_value = decoder.read(json.loads(json_str))
                re_encoded_str = json.dumps(encoder.write(decoded_value))
                re_decoded_value = decoder.read(json.loads(re_encoded_str))
                assert decoded_value == re_decoded_value, "Encoding + decoding =/= identity (for '%s')." % json_str
            except Exception:
                assert False, "Exception while proving encoding + decoding = identity (example: '%s')." % json_str


@dataclass
class C3(common.C):
    c3_specific: int


def test_unknown_subclass() -> None:
    c3 = C3(common=10, c3_specific=20)
    for c_encoder in [c_flat_encoder, c_nested_encoder, c_untagged_encoder]:
        try:
            c_encoder.write(c3)
            assert False, "Unknown class C3 (at the time of encoder extraction) was successfully serialized"
        except JsonEncoderException:
            pass


def test_nested_generics() -> None:
    # The following just tests that extractor does not go into an infinite loop
    auto_json_encoder.extract(G2[C1])
    auto_json_encoder.extract(G[G[C1]])
    auto_json_encoder.extract(G[G[G[C1]]])



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Callable
from typing import cast, Tuple
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import TypeVar
from typing import Union

from pytyped.macros.boxed import Boxed
from pytyped.macros.extractor import Extractor
from pytyped.macros.extractor import WithDefault
from pytyped.json.common import JsValue


@dataclass
class JsonEncoderException(Exception):
    message: str


T = TypeVar("T")


class JsonEncoder(Generic[T], metaclass=ABCMeta):
    is_optional: bool = False

    @abstractmethod
    def encode(self, t: T) -> JsValue:
        pass

    def write(self, t: T) -> JsValue:
        return self.encode(t)


@dataclass
class JsonObjectEncoder(JsonEncoder[T]):
    field_encoders: Dict[str, JsonEncoder[Any]]

    def encode(self, t: T) -> JsValue:
        d: Dict[str, Any]
        if hasattr(t, "_asdict"):
            d = cast(NamedTuple, t)._asdict()
        else:
            d = t.__dict__

        return {
            field_name: field_encoder.encode(d[field_name])
            for field_name, field_encoder in self.field_encoders.items()
        }


@dataclass
class JsonTupleEncoder(JsonEncoder[Tuple[Any, ...]]):
    field_encoders: List[JsonEncoder[Any]]

    def encode(self, t: Tuple[Any, ...]) -> JsValue:
        return [e.encode(v) for (v, e) in zip(t, self.field_encoders)]


@dataclass
class JsonTaggedEncoder(JsonEncoder[T]):
    branches: Dict[str, Tuple[type, JsonEncoder[Any]]]
    tag_field_name: str = "tag"
    value_field_name: Optional[str] = None

    def encode(self, t: T) -> JsValue:
        for branch_name, (branch_type, encoder) in self.branches.items():
            if isinstance(t, branch_type):
                encoded_value: JsValue = encoder.encode(t)
                if self.value_field_name is not None:
                    encoded_value = {self.value_field_name: encoded_value}
                if not isinstance(encoded_value, dict):
                    raise JsonEncoderException("Only JSON objects are supported as encoded value of named sum types.")
                encoded_value[self.tag_field_name] = branch_name
                return encoded_value

        raise JsonEncoderException("Unknown subclass (known subclasses are %s)" % (", ".join(self.branches.keys())))


@dataclass
class JsonPriorityEncoder(JsonEncoder[T]):
    branches: List[Tuple[type, JsonEncoder[Any]]]

    def encode(self, t: T) -> JsValue:
        for branch_type, encoder in self.branches:
            if isinstance(t, branch_type):
                return encoder.encode(t)

        raise JsonEncoderException(
            "Unknown sub-type (known sub-types are %s)" % (", ".join([str(t) for (t, _) in self.branches]))
        )


@dataclass
class JsonOptionalEncoder(JsonEncoder[Optional[T]]):
    inner_encoder: JsonEncoder[T]

    def encode(self, t: Optional[T]) -> JsValue:
        if t is None:
            return None

        return self.inner_encoder.encode(t)


U = TypeVar("U")


@dataclass
class JsonMappedEncoder(JsonEncoder[T], Generic[T, U]):
    u_encoder: JsonEncoder[U]
    t_to_u: Callable[[T], U]

    def encode(self, t: T) -> JsValue:
        transformer = cast(Callable[[T], U], self.t_to_u)
        return self.u_encoder.encode(transformer(t))


@dataclass
class JsonBoxedEncoder(JsonEncoder[T]):
    field_name: str
    field_encoder: JsonEncoder[Any]

    def encode(self, t: T) -> JsValue:
        d = cast(NamedTuple, t)._asdict()
        return self.field_encoder.encode(d[self.field_name])


@dataclass
class JsonListEncoder(JsonEncoder[List[T]]):
    element_encoder: JsonEncoder[T]

    def encode(self, list: List[T]) -> JsValue:
        return [self.element_encoder.encode(i) for i in list]


@dataclass
class JsonStringDictionaryEncoder(JsonEncoder[Dict[str, T]]):
    element_encoder: JsonEncoder[T]

    def encode(self, d: Dict[str, T]) -> JsValue:
        return {k: self.element_encoder.encode(v) for k, v in d.items()}


@dataclass
class JsonBasicEncoder(
    JsonEncoder[Union[str, int, bool, float, Decimal, None]]
):
    def encode(
        self, t: Union[str, int, bool, float, Decimal, None]
    ) -> JsValue:
        return t


@dataclass
class JsonDecimalEncoder(JsonEncoder[Decimal]):
    def encode(self, t: Decimal) -> JsValue:
        return str(t)


@dataclass
class JsonDateEncoder(JsonEncoder[Union[date, datetime]]):
    def encode(self, d: Union[date, datetime]) -> JsValue:
        return d.isoformat()


@dataclass
class JsonEnumEncoder(JsonEncoder[Enum]):
    def encode(self, t: Enum) -> JsValue:
        return str(t.value)


@dataclass
class JsonNoneEncoder(JsonEncoder[None]):
    def encode(self, t: None) -> JsValue:
        return {}


class AutoJsonEncoder(Extractor[JsonEncoder[Any]]):
    json_basic_encoder: JsonBasicEncoder = JsonBasicEncoder()
    json_decimal_encoder: JsonDecimalEncoder = JsonDecimalEncoder()
    json_date_encoder: JsonDateEncoder = JsonDateEncoder()
    json_enum_encoder: JsonEnumEncoder = JsonEnumEncoder()

    basic_encoders: Dict[type, Boxed[JsonEncoder[Any]]] = {
        bool: Boxed(json_basic_encoder),
        str: Boxed(json_basic_encoder),
        int: Boxed(json_basic_encoder),
        float: Boxed(json_basic_encoder),
        Decimal: Boxed(json_decimal_encoder),
        datetime: Boxed(json_date_encoder),
        date: Boxed(json_date_encoder),
        type(None): Boxed(JsonNoneEncoder())
    }

    def __init__(self) -> None:
        super().__init__()

    @property
    def basics(self) -> Dict[type, Boxed[JsonEncoder[Any]]]:
        return self.basic_encoders

    def named_product_extractor(self, t: type, fields: Dict[str, WithDefault[JsonEncoder[Any]]]) -> JsonEncoder[Any]:
        field_encoders: Dict[str, JsonEncoder[Any]] = {
            n: v.t for (n, v) in fields.items()
        }
        return JsonObjectEncoder(field_encoders=field_encoders)

    def unnamed_product_extractor(self, t: type, fields: List[JsonEncoder[Any]]) -> JsonEncoder[Tuple[Any, ...]]:
        return JsonTupleEncoder(fields)

    def named_sum_extractor(self, t: type, branches: Dict[str, Tuple[type, JsonEncoder[Any]]]) -> JsonEncoder[Any]:
        return JsonTaggedEncoder(
            branches=branches,
            tag_field_name=t.__name__,
            value_field_name=None
        )

    def unnamed_sum_extractor(self, t: type, branches: List[Tuple[type, JsonEncoder[Any]]]) -> JsonEncoder[Any]:
        return JsonPriorityEncoder(branches)

    def optional_extractor(self, t: JsonEncoder[T]) -> JsonEncoder[Optional[T]]:
        return JsonOptionalEncoder(t)

    def list_extractor(self, t: JsonEncoder[T]) -> JsonEncoder[List[T]]:
        return JsonListEncoder(t)

    def dictionary_extractor(
        self,
        key: type,
        value: type,
        key_ext: JsonEncoder[Any],
        val_ext: JsonEncoder[Any]
    ) -> JsonEncoder[Dict[Any, Any]]:
        def to_str_dict(d: Dict[Enum, Any]) -> Dict[str, Any]:
            return {cast(str, k.value): v for k, v in d.items()}

        if key is str:
            return JsonStringDictionaryEncoder(val_ext)
        if issubclass(key, Enum):
            return JsonMappedEncoder(JsonStringDictionaryEncoder(val_ext), to_str_dict)
        raise NotImplementedError()

    def enum_extractor(self, enum_name: str, enum_values: List[Tuple[str, Any]]) -> JsonEncoder[Any]:
        return self.json_enum_encoder




from dataclasses import dataclass
from dataclasses import field
from datetime import date
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict
from typing import Generic
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from pytyped.json.decoder import AutoJsonDecoder
from pytyped.json.decoder import JsonBoxedDecoder
from pytyped.json.decoder import JsonDecoder
from pytyped.json.decoder import JsonTaggedDecoder


class E(Enum):
    X = "A"
    Y = "B"
    Z = "C"


@dataclass
class A:
    d: Optional[date]
    dt: Optional[datetime]
    e: Optional[E]
    x: int
    y: bool
    dec: Decimal = Decimal("10.123")
    t: Tuple[str, int] = ("t_str", 10)
    z: str = "abc"


@dataclass
class B:
    maybe_a: Optional[A]
    a_list: List[A] = field(default_factory=lambda: [A(
        date(2019, 1, 1),
        datetime(2019, 1, 1, 0, 0, 0, 0, None),
        E.Z,
        10,
        True,
        z="List[A]",
    )])


class BoxedA(NamedTuple):
    value: A


@dataclass
class C:
    common: int


@dataclass
class C1(C):
    c1_specific: str


@dataclass
class C2(C):
    c2_specific: str


@dataclass
class D:
    common: int
    c1_specific: Optional[str]
    c2_specific: Optional[str]


T = TypeVar("T")


@dataclass
class G(Generic[T]):
    non_generic_field: str
    generic_field: T


@dataclass
class G2(Generic[T]):
    gs: List[G[T]]


@dataclass
class Composite:
    some_field: int
    int_g: G[int]
    str_g: G[str]
    int_list_g: G[List[int]]


valid_a_jsons = [
    '{"x": 1, "y": false, "z": "xyz"}',
    '{"x": 1, "y": false, "t": ["abc", 8]}',
    '{"x": 1, "y": false}',
    '{"x": "1", "y": false}',
    '{"x": 1, "y": false, "z": null}',
    '{"e": "A", "x": 1, "y": false}',
    '{"d": "2019-01-20", "x": 1, "y": false}',
    '{"dt": "2019-01-20T11:11:11", "x": 1, "y": false}',
]

a_list_json = "[" + (", ".join(valid_a_jsons)) + "]"

b_with_a_json_possibilities = (
        [
            '{"maybe_a": ' + a_possibility + "}"
            for a_possibility in valid_a_jsons
        ]
        + [
            '{"maybe_a": ' + a_possibility + ', "a_list": null}'
            for a_possibility in valid_a_jsons
        ]
        + [
            '{"maybe_a": ' + a_possibility + ', "a_list": ' + a_list_json + "}"
            for a_possibility in valid_a_jsons
        ]
)
b_without_a_json_possibilities = [
    "{}",
    '{"maybe_a": null, "a_list": null}',
    '{"a_list": ' + a_list_json + "}",
    '{"maybe_a": null, "a_list": ' + a_list_json + "}",
]
valid_b_jsons: List[str] = b_with_a_json_possibilities + b_without_a_json_possibilities

c_valid_flat_json_strs: List[str] = [
    '{"common": 10, "c1_specific": "abc", "C": "C1"}',
    '{"common": 10, "c2_specific": "def", "C": "C2"}'
]
c_valid_nested_json_strs: List[str] = [
    '{"value": {"common": 10, "c1_specific": "abc"}, "tag": "C1"}',
    '{"value": {"common": 10, "c2_specific": "def"}, "tag": "C2"}'
]
c_valid_untagged_json_strs: List[str] = [
    '{"common": 10, "c1_specific": "abc"}',
    '{"common": 10, "c2_specific": "def"}'
]

valid_string_dictionary_jsons: List[str] = [
    '{}',
    '{"a": 1, "b": 2}',
    '{"c": 1, "b": 2}',
]

valid_composite_jsons: List[str] = [
    """{
    "some_field": 10,
    "int_g": {"non_generic_field": "abc", "generic_field": 17},
    "str_g": {"non_generic_field": "def", "generic_field": "some_string"},
    "int_list_g": {"non_generic_field": "ghi", "generic_field": [10, 11, 12]}
}"""
]

auto_json_decoder = AutoJsonDecoder()

a_decoder = auto_json_decoder.extract(A)
b_decoder = auto_json_decoder.extract(B)
boxed_a_decoder = JsonBoxedDecoder("value", a_decoder, lambda d: BoxedA(**d))

c_flat_decoder = auto_json_decoder.extract(C)
c_nested_decoder = cast(
    JsonDecoder[C],
    JsonTaggedDecoder(
        branch_decoders={
            "C1": auto_json_decoder.extract(C1),
            "C2": auto_json_decoder.extract(C2)
        },
        tag_field_name="tag",
        value_field_name="value"
    )
)
c_untagged_decoder = cast(JsonDecoder[C], auto_json_decoder.extract(cast(type, Union[C1, C2])))

string_to_int_dic_json_decoder = auto_json_decoder.extract(Dict[str, int])

composite_decoder = cast(JsonDecoder[Composite], auto_json_decoder.extract(Composite))



from decimal import Decimal
from typing import Any
from typing import Dict
from typing import List
from typing import Union

JsValue = Union[Dict[str, Any], List[Any], str, Decimal, float, int, bool, None]



import json
from typing import Dict
from typing import Tuple
from typing import Union
from typing import cast

from typing import List, Optional

from pytyped.json.decoder import JsDecodeException
from pytyped.json.decoder import JsDecodeErrorInArray
from pytyped.json.decoder import JsDecodeErrorInField
from pytyped.json.decoder import JsDecodeErrorFinal
from pytyped.json.decoder import JsonDecoder
from pytyped.json.decoder import JsonErrorAsDefaultDecoder
from pytyped.json.decoder import JsonMappedDecoder
from pytyped.json.decoder import JsonTaggedDecoder
from tests.json import common
from tests.json.common import C1
from tests.json.common import G
from tests.json.common import G2
from tests.json.common import auto_json_decoder

default_a = common.A(d=None, dt=None, e=None, x=100, y=False, t=("t_str2", 20), z="default")
a_with_default_decoder = JsonErrorAsDefaultDecoder(common.a_decoder, default_a)

invalid_a_jsons = [
    '{"x": [], "y": false}',  # expected integer, received JsArray
    '{"x": 1.5, "y": false}',  # non-integral integer
    '{"x": 1, "y": "string"}',  # Boolean expected but string found
    '{"x": 1, "z": "abc"}',  # Required field missing
    '{"x": 1, "y": false, "z": 1}',  # Expected string but received numeric
    '{"x": 1, "y": false, "t": "abc"}',  # A tuple is supposed to be encoded as a JsArray.
    '{"x": 1, "y": false, "t": {}}',  # A tuple is supposed to be encoded as a JsArray.
    '{"x": 1, "y": false, "t": ["abc"]}',  # Wrong number of arguments for tuple
    '{"x": 1, "y": false, "t": ["abc", 10, "efg"]}',  # Wrong number of arguments for tuple
    '{"x": 1, "y": false, "t": [1, 10]}',  # Wrong type of argument in tuple
    '{"x": 1, "y": false, "t": ["abc", "def"]}',  # Wrong type of argument in tuple
    '{"e": 1, "x": 1, "y": false}',  # Expected enum E but received numeric
    '{"e": "X", "x": 1, "y": false}',  # Expected enum E but received string that cannot be decoded into date
    '{"d": 1, "x": 1, "y": false}',  # Expected date but received numeric
    '{"d": "abc", "x": 1, "y": false}',  # Expected date but received string that cannot be decoded into date
    '{"dt": 1, "x": 1, "y": false}',  # Expected datetime but received numeric
    '{"dt": "abc", "x": 1, "y": false}',  # Expected datetime but received string that cannot be decoded into date
    '[{"x": 1, "z": "abc"}]',  # Expected JsObject but received JsArray
]

invalid_b_jsons_set_1 = [
    '{"maybe_a": ' + invalid_a_json + "}" for invalid_a_json in invalid_a_jsons
]
invalid_b_jsons_set_2 = [
    '{"a_list": [' + invalid_a_json + "]}"
    for invalid_a_json in invalid_a_jsons
]
invalid_b_jsons_set_3 = [
    '{"a_list": ' + valid_a_json + "}" for valid_a_json in common.valid_a_jsons
]


def test_json_decoder_valid() -> None:
    for a_json in common.valid_a_jsons:
        try:
            decoded_a = common.a_decoder.read(json.loads(a_json))
            assert isinstance(decoded_a, common.A)
            assert decoded_a.x == 1
            assert not decoded_a.y
            assert decoded_a.t[0] in ["t_str", "abc"]
            assert decoded_a.t[1] in [8, 10]
            assert decoded_a.z in ["abc", "xyz"]
        except Exception as e:
            assert False, (
                "Caught an exception while decoding valid JSONs. Exception: %s"
                % str(e)
            )

    for b_json in common.b_with_a_json_possibilities:
        try:
            decoded_b = common.b_decoder.read(json.loads(b_json))
            assert isinstance(decoded_b, common.B)
            assert len(decoded_b.a_list) in {1, len(common.valid_a_jsons)}
            if len(decoded_b.a_list) == 1:
                assert decoded_b.a_list[0].x == 10
                assert decoded_b.a_list[0].y
            else:
                assert decoded_b.a_list[0].x == 1
                assert not decoded_b.a_list[0].y
            assert decoded_b.maybe_a is not None
            assert decoded_b.maybe_a.x == 1
            assert not decoded_b.maybe_a.y
            assert (decoded_b.maybe_a.z == "abc") or (
                decoded_b.maybe_a.z == "xyz"
            )
        except Exception as e:
            assert False, (
                "Caught an exception while decoding valid JSONs. Exception: %s"
                % str(e)
            )

    for b_json in common.b_without_a_json_possibilities:
        try:
            decoded_b = common.b_decoder.read(json.loads(b_json))
            assert isinstance(decoded_b, common.B)
            if len(decoded_b.a_list) == 1:
                assert decoded_b.a_list[0].x == 10
                assert decoded_b.a_list[0].y
            else:
                assert decoded_b.a_list[0].x == 1
                assert not decoded_b.a_list[0].y
            assert decoded_b.maybe_a is None
        except Exception as e:
            assert False, (
                "Caught an exception while decoding valid JSONs. Exception: %s"
                % str(e)
            )


def test_json_decoder_invalid() -> None:
    for invalid_a_json in invalid_a_jsons:
        try:
            x = common.a_decoder.read(json.loads(invalid_a_json))
            assert False, "Invalid JSON %s was successfully decoded as %s!" % (
                invalid_a_json,
                str(x),
            )
        except JsDecodeException as e:
            assert len(e.errors) == 1

    for invalid_b_json in invalid_b_jsons_set_1:
        try:
            _ = common.b_decoder.read(json.loads(invalid_b_json))
            assert False, "Invalid JSON was successfully decoded!"
        except JsDecodeException as e:
            assert len(e.errors) == 1
            assert isinstance(e.errors[0], JsDecodeErrorInField)

    for invalid_b_json in invalid_b_jsons_set_2:
        try:
            _ = common.b_decoder.read(json.loads(invalid_b_json))
            assert False, "Invalid JSON was successfully decoded!"
        except JsDecodeException as e:
            assert len(e.errors) == 1
            assert str(e).startswith("Found 1 error")
            error = e.errors[0]
            assert isinstance(error, JsDecodeErrorInField)
            assert isinstance(error.error, JsDecodeErrorInArray)

    for invalid_b_json in invalid_b_jsons_set_3:
        try:
            _ = common.b_decoder.read(json.loads(invalid_b_json))
            assert False, "Invalid JSON was successfully decoded!"
        except JsDecodeException as e:
            assert len(e.errors) == 1
            assert str(e).startswith("Found 1 error")
            error = e.errors[0]
            assert isinstance(error, JsDecodeErrorInField)
            assert isinstance(error.error, JsDecodeErrorFinal)


def test_boxed_decoder() -> None:
    for a_json in common.valid_a_jsons:
        try:
            decoded_boxed_a = common.boxed_a_decoder.read(json.loads(a_json))
            assert isinstance(decoded_boxed_a, common.BoxedA)
            assert decoded_boxed_a.value.x == 1
            assert not decoded_boxed_a.value.y
            assert (decoded_boxed_a.value.z == "abc") or (
                decoded_boxed_a.value.z == "xyz"
            )
        except Exception as e:
            assert False, (
                "Caught an exception while decoding valid JSONs. Exception: %s"
                % str(e)
            )

    for invalid_a_json in invalid_a_jsons:
        try:
            _ = common.boxed_a_decoder.read(json.loads(invalid_a_json))
            assert False, "Invalid JSON was successfully decoded!"
        except JsDecodeException as e:
            assert len(e.errors) == 1


def test_error_as_default_decoder() -> None:
    for a_json in common.valid_a_jsons:
        a = a_with_default_decoder.read(json.loads(a_json))
        assert a.x == 1
        assert not a.y
        assert (a.z == "abc") or (a.z == "xyz")

    for invalid_a_json in invalid_a_jsons:
        a = a_with_default_decoder.read(json.loads(invalid_a_json))
        assert a == default_a


def test_optional_decoder_none() -> None:
    assert common.auto_json_decoder.extract(cast(type, Optional[common.A])).read(json.loads("null")) is None


def valid_c_test_cases(decoder: JsonDecoder[common.C], valid_jsons: List[str]) -> None:
    for s in valid_jsons:
        try:
            c: common.C = decoder.read(json.loads(s))
            assert c.common == 10
            if isinstance(c, common.C1):
                assert c.c1_specific == "abc"
            else:
                assert isinstance(c, common.C2)
                assert c.c2_specific == "def"
        except JsDecodeException:
            assert False, "Valid JSON failed to be decoded: '%s'." % s


def invalid_c_test_cases(decoder: JsonDecoder[common.C], invalid_jsons: List[str]) -> None:
    for s in invalid_jsons:
        try:
            decoder.read(json.loads(s))
            assert False, "Invalid JSON was decoded successfully: '%s'." % s
        except JsDecodeException:
            pass


def valid_d_test_cases(decoder: JsonDecoder[common.D], valid_jsons: List[str]) -> None:
    for s in valid_jsons:
        try:
            d: common.D = decoder.read(json.loads(s))
            assert d.common == 10
            assert d.c1_specific in [None, "abc"]
            assert d.c2_specific in [None, "def"]
            assert None in [d.c1_specific, d.c2_specific]
        except JsDecodeException:
            assert False, "Valid JSON failed to be decoded: '%s'." % s


def invalid_d_test_cases(decoder: JsonDecoder[common.D], invalid_jsons: List[str]) -> None:
    for s in invalid_jsons:
        try:
            decoder.read(json.loads(s))
            assert False, "Invalid JSON was decoded successfully: '%s'." % s
        except JsDecodeException:
            pass


c_invalid_flat_json_strs: List[str] = [
    '[]',  # named unions are expected to be JsObjects
    '10',  # named unions are expected to be JsObjects
    'null',  # named unions are expected to be JsObjects
    '"C1"',  # named unions are expected to be JsObjects
    '[{"common": 10, "c1_specific": "abc", "C": "C1"}]',  # named unions are expected to be JsObjects
    '{"common": 10, "c1_specific": "abc", "C": 10}',  # tag is supposed to be a string
    '{"common": 10, "c1_specific": "abc"}',  # tag is missing
    '{"common": 10, "c1_specific": "def", "C": "C3"}',  # unknown tag
    '{"common": 10, "c2_specific": "def", "C": "C3"}',  # unknown tag
    '{"common": "abc", "c1_specific": "def", "C": "C1"}',  # expected integer but received string
    '{"common": 10, "c1_specific": 10, "C": "C1"}',  # expected string but received integer
    '{"common": "abc", "c2_specific": "def", "C": "C2"}',  # expected integer but received string
    '{"common": 10, "c2_specific": 10, "C": "C2"}',  # expected string but received integer
    '{"common": "abc", "c2_specific": "def", "C": "C1"}',  # incorrect tag
    '{"common": "abc", "c1_specific": "def", "C": "C2"}',  # incorrect tag
]

c_invalid_nested_json_strs: List[str] = [
    '[]',  # named unions are expected to be JsObjects
    '10',  # named unions are expected to be JsObjects
    'null',  # named unions are expected to be JsObjects
    '"C1"',  # named unions are expected to be JsObjects
    '[{"common": 10, "c1_specific": "abc", "tag": "C1"}]',  # named unions are expected to be JsObjects
    '{"common": 10, "c1_specific": "abc", "tag": 10}',  # tag is supposed to be a string
    '{"common": 10, "c1_specific": "abc", "tag": "C1"}',  # flat encoding instead of nested encoding
    '{"value": {"common": 10, "c1_specific": "abc"}}',  # tag is missing
    '{"value": {"common": 10, "c1_specific": "def"}, "tag": "C3"}',  # unknown tag
    '{"value": {"common": 10, "c2_specific": "def"}, "tag": "C3"}',  # unknown tag
    '{"value": {"common": "abc", "c1_specific": "def"}, "tag": "C1"}',  # expected integer but received string
    '{"value": {"common": 10, "c1_specific": 10}, "tag": "C1"}',  # expected string but received integer
    '{"value": {"common": "abc", "c2_specific": "def"}, "tag": "C2"}',  # expected integer but received string
    '{"value": {"common": 10, "c2_specific": 10}, "tag": "C2"}',  # expected string but received integer
    '{"value": {"common": "abc", "c2_specific": "def"}, "tag": "C1"}',  # incorrect tag
    '{"value": {"common": "abc", "c1_specific": "def"}, "tag": "C2"}',  # incorrect tag
]

c_invalid_untagged_json_strs: List[str] = [
    '[]',  # not decodable as either C1 or C2
    '10',  # not decodable as either C1 or C2
    'null',  # not decodable as either C1 or C2
    '"C1"',  # not decodable as either C1 or C2
    '[{"common": 10, "c1_specific": "abc"}]',  # not decodable as either C1 or C2
    '{"common": "abc", "c1_specific": "def"}',  # expected integer but received string
    '{"common": 10, "c1_specific": 10}',  # expected string but received integer
    '{"common": "abc", "c2_specific": "def"}',  # expected integer but received string
    '{"common": 10, "c2_specific": 10}'  # expected string but received integer
]


c_decoding_tests: List[Tuple[JsonDecoder[common.C], List[str], List[str]]] = [
    (common.c_flat_decoder, common.c_valid_flat_json_strs, c_invalid_flat_json_strs),
    (common.c_nested_decoder, common.c_valid_nested_json_strs, c_invalid_nested_json_strs),
    (common.c_untagged_decoder, common.c_valid_untagged_json_strs, c_invalid_untagged_json_strs)
]


def test_c_decoders() -> None:
    for decoder, valid_jsons, invalid_jsons in c_decoding_tests:
        valid_c_test_cases(decoder, valid_jsons)
        invalid_c_test_cases(decoder, invalid_jsons)


def c_to_d(c: common.C) -> common.D:
    return common.D(
        common=c.common,
        c1_specific=c.c1_specific if isinstance(c, common.C1) else None,
        c2_specific=c.c2_specific if isinstance(c, common.C2) else None
    )


d_decoding_tests: List[Tuple[JsonDecoder[common.D], List[str], List[str]]] = [
    (JsonMappedDecoder(decoder, c_to_d), valids, invalids) for (decoder, valids, invalids) in c_decoding_tests
]


def test_d_decoding() -> None:
    for decoder, valid_jsons, invalid_jsons in d_decoding_tests:
        valid_d_test_cases(decoder, valid_jsons)
        invalid_d_test_cases(decoder, invalid_jsons)


def test_string_dictionary_decoder() -> None:
    invalid_json_strs: List[str] = [
        '[]',
        'null',
        '10',
        '"abc"',
        '[{}]',
        '{"a": "abc", "b": 2}'
    ]

    for json_str in common.valid_string_dictionary_jsons:
        try:
            common.string_to_int_dic_json_decoder.read(json.loads(json_str))
        except JsDecodeException:
            assert False, "Valid JSON failed to be deserialized: '%s'." % json_str

    for json_str in invalid_json_strs:
        try:
            common.string_to_int_dic_json_decoder.read(json.loads(json_str))
            assert False, "Invalid JSON was successfully deserialized: '%s'." % json_str
        except JsDecodeException:
            pass


def test_composite_decoder() -> None:
    for json_str in common.valid_composite_jsons:
        composite = common.composite_decoder.read(json.loads(json_str))
        assert isinstance(composite.int_g.generic_field, int)
        assert isinstance(composite.str_g.generic_field, str)
        assert isinstance(composite.int_list_g.generic_field, list)
        for n in composite.int_list_g.generic_field:
            assert isinstance(n, int)


def test_nested_generics() -> None:
    # The following just tests that extractor does not go into an infinite loop
    auto_json_decoder.extract(G2[C1])
    auto_json_decoder.extract(G[G[C1]])
    auto_json_decoder.extract(G[G[G[C1]]])
